"""Access station information and status. More to come."""

from lyft_bikes.live import LiveRequest


class StationBase(LiveRequest):
    key: str = "stations"


class StationInfo(StationBase):
    url: str = "https://gbfs.divvybikes.com/gbfs/en/station_information.json"


class StationStatus(StationBase):
    url: str = "https://gbfs.divvybikes.com/gbfs/en/station_status.json"



import datetime
import io
import requests
import zipfile

import pandas as pd

from lyft_bikes.live import BadRequest


class BaseDownloader:
    def __init__(self, http_client=requests) -> None:
        self.http_client = http_client

    def file_name(self, date: datetime.date, suffix: str) -> str:
        raise NotImplementedError(
            "This method should be implemented in the child class."
        )

    @property
    def base_url(self) -> str:
        raise NotImplementedError(
            "This method should be implemented in the child class."
        )

    def url(self, date: datetime.date):
        return f"{self.base_url}/{self.file_name(date=date, suffix='zip')}"

    def read(self, date: datetime.date) -> pd.DataFrame:
        url = self.url(date=date)
        response = self.http_client.get(url)

        if not response.ok:
            raise BadRequest(f"The response for {url} wasn't okay.")

        zipdata = zipfile.ZipFile(io.BytesIO(response.content))

        return pd.read_csv(zipdata.open(self.file_name(date=date, suffix="csv")))


class DivvyDownloader(BaseDownloader):
    """Class to download historical trips from Divvy in Chicago.

    Index for all the historical trips found <a href="https://divvy-tripdata.s3.amazonaws.com/index.html">here</a>.

    Currently only supports the files with the form `%Y%m-divvy-tripdata.zip` that go back until
    April 2020

    """

    base_url = "https://divvy-tripdata.s3.amazonaws.com"

    def file_name(self, date: datetime.date, suffix: str) -> str:
        return f"{date:%Y%m}-divvy-tripdata.{suffix}"


class CitiBikesDownloader(BaseDownloader):
    """Class to download historical trips from CitiBikes in New York City.

    Index for all the historical trips found <a href="https://s3.amazonaws.com/tripdata/index.html">here</a>.

    """

    base_url = "https://s3.amazonaws.com/tripdata"

    def file_name(self, date: datetime.date, suffix: str) -> str:
        return f"JC-{date:%Y%m}-citibike-tripdata.{suffix}"

    def url(self, date: datetime.date):
        return f"{self.base_url}/{self.file_name(date=date, suffix='csv.zip')}"


class BayWheelsDownloader(BaseDownloader):
    """Class to download historical trips from BayBikes in San Francisco.

    Index for all the historical trips found <a href="https://s3.amazonaws.com/baywheels-data/index.html">here</a>.

    """

    base_url = "https://s3.amazonaws.com/baywheels-data"

    def file_name(self, date: datetime.date, suffix: str) -> str:
        return f"{date:%Y%m}-baywheels-tripdata.{suffix}"

    def url(self, date: datetime.date):
        return f"{self.base_url}/{self.file_name(date=date, suffix='csv.zip')}"


class CoGoDownloader(BaseDownloader):
    """Class to download historical trips from CoGo in Columbus.

    Index for all the historical trips found <a href="https://cogo-sys-data.s3.amazonaws.com/index.html">here</a>.

    """

    base_url = "https://cogo-sys-data.s3.amazonaws.com"

    def file_name(self, date: datetime.date, suffix: str) -> str:
        return f"{date:%Y%m}-cogo-tripdata.{suffix}"


class CapitalBikeshareDownloader(BaseDownloader):
    """Class to download historical trips from Capital Bikeshare in Washington DC.

    Index for all the historical trips found <a href="https://s3.amazonaws.com/capitalbikeshare-data/index.html">here</a>.

    """

    base_url = "https://s3.amazonaws.com/capitalbikeshare-data"

    def file_name(self, date: datetime.date, suffix: str) -> str:
        return f"{date:%Y%m}-capitalbikeshare-tripdata.{suffix}"



import pandas as pd

from typing import Dict

import requests
import json

import datetime


class BadRequest(Exception):
    """Some error happened with the request."""


class RequestBase:
    """Base for getting data sources from Divvy gbfs.

    Get a JSON data source from url and cache it

    """

    def __init__(self) -> None:
        self._data = None

    def data(self, url: str) -> Dict:
        if self._data is not None:
            return self._data

        response = requests.get(url)

        if not response.ok:
            raise BadRequest(f"The response to {url} was not ok. {response.text}")

        return json.loads(response.text)


class LiveRequest(RequestBase):
    key: str = None
    url: str = None

    def __init__(self, requests_base: RequestBase = None) -> None:
        if requests_base is None:
            requests_base = RequestBase()

        self.requests_base = requests_base

    @property
    def data(self) -> Dict:
        return self.requests_base.data(self.url)

    @property
    def last_updated(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self.data["last_updated"])

    def read(self) -> pd.DataFrame:
        return pd.DataFrame(self.data["data"][self.key])



"""Access the live feed of bikes and scooters. More to come."""

from lyft_bikes.live import LiveRequest


class Bikes(LiveRequest):
    """Fetch the live feed of bikes and scooters.

    Example:
        Get the live ebikes and scooters as pandas.DataFrame

        ```python
        live_bikes = Bikes()
        df_live_bikes = live_bikes.read()
        ```

    """

    url: str = "https://gbfs.divvybikes.com/gbfs/en/free_bike_status.json"
    key: str = "bikes"



from pathlib import Path

try:
    import geopandas as gpd
except ImportError:
    gpd = None


def read_fee_boundary() -> "GeoDataFrame":
    """Read in the fee boundary before the pricing change in May 2022.

    Polygon created by referencing the divvy site and using online software
    to create custom map: https://www.keene.edu/campus/maps/tool/

    Returns:
        GeoDataFrame with single row geometry

    """
    if gpd is None:
        msg = "The package 'geopandas' must be installed to read fee boundary."
        raise ImportError(msg)

    file = Path(__file__).parent / "fees.geojson"

    return gpd.read_file(file, crs="EPSG:4326")



from datetime import date
from typing import Protocol, Union
import warnings

import pandas as pd


class Dates(Protocol):
    @property
    def last_date(self) -> date: ...

    def create_date_range(self, start_date: str, end_date: str) -> pd.Series: ...

    def to_date(self, year: int, month: int) -> date: ...

    def check_valid(self, date: date) -> None: ...


class Downloader(Protocol):
    def read(self, date: date) -> pd.DataFrame: ...


class HistoricalTrips:
    def __init__(self, dates: Dates, downloader: Downloader):
        self.dates = dates
        self.downloader = downloader

    def get_months_trips(self, year: int, month: int) -> pd.DataFrame:
        """Return pandas.DataFrame for a given year and month"""
        date = self.dates.to_date(year, month)
        self.dates.check_valid(date)

        return self.downloader.read(date=date)

    def read(
        self,
        start_date: str,
        start_time_col: str = "started_at",
        end_date: Union[str, None] = None,
    ) -> pd.DataFrame:
        """Return historical trips for a given range of dates

        Args:
            start_date: start date for the data in %Y-%m-%d format
            start_col_col: column name for the start time
            end_date: end date in the same format. Defaults to last date available

        Returns:
            Historical trip DataFrame for the date range provided.

        Examples:
            Read trips from Jan 1st 2021 until Feb 1st 2021

            >>> trips = HistoricalTrips()
            >>> df_trips = trips.read(start_date="2021-01-01", end_date="2021-02-01")

            Read trips from Jan 1st 2021 until the last date available

            >>> df_trips = trips.read(start_date="2021-01-01")

        """
        if end_date is None:
            end_date = str(self.dates.last_date)

        df_trips = pd.concat(
            [
                self.get_months_trips(date.year, date.month)
                for date in self.dates.create_date_range(start_date, end_date)
            ],
            ignore_index=True,
        )

        try:
            date = pd.to_datetime(df_trips[start_time_col]).dt.date.astype(str)
            idx = (date >= start_date) & (date <= end_date)
            return df_trips.loc[idx, :].reset_index(drop=True)
        except KeyError:
            warnings.warn(
                f"The column {start_time_col} does not exist. Returning all trips instead.",
                UserWarning,
            )
            return df_trips



import pandas as pd

from functools import wraps
from typing import Dict, Union

from lyft_bikes.bikes import Bikes
from lyft_bikes.stations import StationInfo, StationStatus

from lyft_bikes.historical.historical import HistoricalTrips
from lyft_bikes.historical import dates as bike_dates, downloader

from lyft_bikes.geo import read_fee_boundary

__version__ = "0.1.3"


def chicago_only_error(func) -> None:
    @wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs.get("city") != "Chicago" or args[0] != "Chicago":
            raise ValueError(f"{func.__name__} is only available for Chicago.")

        return func(*args, **kwargs)

    return wrapper


@chicago_only_error
def read_live_bikes(city: str) -> pd.DataFrame:
    """Read currently available bikes and scooters.

    Args:
        city: city to read data from

    Returns:
        DataFrame with the current status of bikes and scooters.

    """
    live = Bikes()
    return live.read()


@chicago_only_error
def read_stations(city: str) -> pd.DataFrame:
    """Read information and status for each station.

    Args:
        city: city to read data from

    Returns:
        DataFrame with the stations information and status.

    """
    station_info = StationInfo()
    station_status = StationStatus()

    return pd.merge(
        station_status.read(),
        station_info.read(),
        how="inner",
        on=["station_id", "legacy_id"],
    )


CITY_DATES: Dict[str, bike_dates.DefaultDates] = {
    "Chicago": bike_dates.DivvyDates,
}


CITY_DOWNLOADERS: Dict[str, downloader.BaseDownloader] = {
    "Chicago": downloader.DivvyDownloader,
    "New York City": downloader.CitiBikesDownloader,
    "San Francisco": downloader.BayWheelsDownloader,
    "Washington DC": downloader.CapitalBikeshareDownloader,
    "Columbus": downloader.CoGoDownloader,
}


def read_historical_trips(
    start_date: str,
    city: str,
    end_date: Union[str, None] = None,
) -> pd.DataFrame:
    """Read historical trips within given range of dates for a city.

    Args:
        start_date: start date for the data in %Y-%m-%d format
        city: city to read data from
        end_date: end date in the same format. Defaults to last date available

    Returns:
        Historical trip DataFrame for the date range provided.

    Examples:
        Read trips from Jan 1st 2021 until Feb 1st 2021 in Chicago

        ```python
        import lyft_bikes

        df = lyft_bikes.read_historical_trips(
            start_date="2021-01-01",
            end_date="2021-02-01",
            city="Chicago",
        )
        ```


    """
    if city not in CITY_DOWNLOADERS:
        raise ValueError(
            f"City {city} is not supported. Supported cities are {CITY_DOWNLOADERS.keys()}"
        )

    dates = CITY_DATES.get(city, bike_dates.DefaultDates)()

    downloader = CITY_DOWNLOADERS[city]()
    trips = HistoricalTrips(
        dates=dates,
        downloader=downloader,
    )
    return trips.read(start_date=start_date, end_date=end_date)



"""Module to calculate pricing of divvy trips."""

import numpy as np
from dataclasses import dataclass


class Rate:
    """All pricing will derive from this class."""

    def __call__(self, duration: np.ndarray) -> np.ndarray:
        """Calculate the rate for a given set of durations. Vectorized."""

    def __add__(self, other):
        return AdditiveRate(self, other)


class AdditiveRate(Rate):
    """A combination of two classes with addition."""

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __call__(self, duration: np.ndarray) -> np.ndarray:
        return self.left(duration=duration) + self.right(duration=duration)

    def __repr__(self) -> str:
        return f"AdditiveRate(left={self.left}, right={self.right})"


@dataclass
class MinuteRate(Rate):
    """Amount in cents per minute starting at `start` minute.

    Args:
        amount: the cost in cents per minute
        start: the minute where that cost starts

    Example:
        Rides that cost 15 cents per minute after 30 minutes of riding.

        ```python
        rate = MinuteRate(amount=15, start=30)
        rate([10, 15, 31]) # [0, 0, 16]
        ```

    """

    amount: int
    start: int

    def __call__(self, duration: np.ndarray) -> np.ndarray:
        return np.maximum(np.rint(duration) - self.start, 0) * self.amount


@dataclass
class UnlockFee(Rate):
    """Amount in cents to unlock the bike."""

    amount: int

    def __call__(self, duration: np.ndarray) -> np.ndarray:
        return np.ones_like(duration) * self.amount



import pandas as pd

from typing import List

import datetime
from calendar import monthrange
from dateutil.relativedelta import relativedelta


class DateRangeError(Exception):
    """This date is before the Divvy start."""


class DefaultDates:
    def check_valid(self, date: datetime.date) -> None:
        if date > self.last_date:
            raise DateRangeError(f"{date} data hasn't been released yet.")

    @property
    def last_date(self) -> datetime.date:
        """Latest available day in the historical dataset.

        Based on the releases of the historical trips. Seems to be safe to include
        the previous month.

        """
        previous_month = datetime.date.today().replace(day=1) - relativedelta(months=1)
        last_day_previous_month = monthrange(previous_month.year, previous_month.month)[
            1
        ]

        return previous_month.replace(day=last_day_previous_month)

    def create_date_range(self, start_date: str, end_date: str) -> List[datetime.date]:
        """Return list of dates from start to end"""
        dates = pd.date_range(
            str(self.first_of_month(start_date)),
            str(self.first_of_month(end_date)),
            freq="MS",
        )

        return [date.to_pydatetime().date() for date in dates]

    @staticmethod
    def first_of_month(date: str) -> datetime.date:
        return datetime.datetime.strptime(date, "%Y-%m-%d").date().replace(day=1)

    @staticmethod
    def to_date(year: int, month: int) -> datetime.date:
        return datetime.date(year, month, 1)


class DivvyDates(DefaultDates):
    """Class that stores and calculates different dates associated with the Divvy Program."""

    first_date = datetime.date(2020, 4, 1)
    electric_trials = datetime.date(2020, 7, 13)
    first_electric_date = datetime.date(2020, 7, 28)
    # Last day of the west of Western pricing
    end_of_waiver = datetime.date(2022, 5, 10)

    def check_valid(self, date: datetime.date) -> None:
        super().check_valid(date)

        if date < self.first_date:
            raise DateRangeError(f"{date} is before the historical trips.")
